import React, { FC, useState } from "react";
import Container from "@material-ui/core/Container";
import Grid from "@material-ui/core/Grid";
import Square from "./Square";
import PlayerIcon from "./PlayerIcon";
import Status from "./Status";

// Helper function for creating tuple types
const MakeTuple = <T extends number[]>(...args: T): T => args;

// Contains the tuple value
const IndexValues = MakeTuple(0, 1, 2, 3, 4, 5, 6, 7, 8);

// Creates a union type from provided tuple
//   the [number] part says the type to acces member of this type with a numbe key
//   which for the tuple contains either on of it's members
type Indexes = typeof IndexValues[number];

export type Player = "O" | "X";

type BoardState = Partial<Record<Indexes, Player>>;

const getWinner = (board: BoardState): Player | undefined => {
    // For board of size 3x3 there are only 8 winning combinations
    const lines = [
        [0, 1, 2],
        [3, 4, 5],
        [6, 7, 8],
        [0, 3, 6],
        [1, 4, 7],
        [2, 5, 8],
        [0, 4, 8],
        [2, 4, 6],
    ] as const;

    // Check all combinations until winner is found
    for (let i = 0; i < lines.length; i++) {
        const [a, b, c] = lines[i];
        if (board[a] && board[a] === board[b] && board[a] === board[c]) {
            return board[a];
        }
    }

    // Winner is no one
    return undefined;
};

const Board: FC = () => {
    // State
    const [player, setPlayer] = useState<Player>("X");
    const [board, setBoard] = useState<BoardState>({});
    const [winner, setWinner] = useState<Player>();

    // Handlers
    const onSquareClicked = (index: Indexes) => {
        // Disallow clicking on already set square
        if (board[index]) {
            return;
        }

        const newBoard = { ...board, [index]: player };

        // Mark square with current player's symbol
        setBoard(newBoard);

        // Switch to other player
        setPlayer((p) => (p === "O" ? "X" : "O"));

        // Update winner
        setWinner(getWinner(newBoard));
    };

    const onBoardRestart = () => {
        setPlayer("X");
        setBoard({});
        setWinner(undefined);
    };

    return (
        <Container maxWidth="sm">
            <Grid container direction="column" spacing={3}>
                <Grid item container spacing={1}>
                    {IndexValues.map((i) => (
                        <Square onClick={() => onSquareClicked(i)}>
                            <PlayerIcon>{board[i] ?? i}</PlayerIcon>
                        </Square>
                    ))}
                </Grid>
                <Status
                    player={player}
                    winner={winner}
                    onRestart={onBoardRestart}
                />
            </Grid>
        </Container>
    );
};

export default Board;
